******************************************************************************************
* Modified version of the graph library from Rabhi and Lapalme's "Algorithms: A Functional Programming Approach"
*
* Sections titled "Sparsity" and "Find Paths" added by Matt Munz
******************************************************************************************

> module Graph(Graph,mkGraph,adjacent,nodes,edgesU,edgesD,edgeIn, weight, findLWPath, findHWPath) where
> import Array
> import List

Graph ADT

Show instance req'd. for arg to weight for debugging purposes

> mkGraph  :: (Ix n,Num w) => Bool -> (n,n) -> [(n,n,w)] -> (Graph n w)
> adjacent :: (Ix n,Num w) => (Graph n w) -> n -> [n]
> nodes    :: (Ix n,Num w) => (Graph n w) -> [n]
> edgesU   :: (Ix n,Num w) => (Graph n w) -> [(n,n,w)]
> edgesD   :: (Ix n,Num w) => (Graph n w) -> [(n,n,w)]
> edgeIn   :: (Ix n,Num w) => (Graph n w) -> (n,n) -> Bool
> weight   :: (Show n,Ix n,Num w) => n -> n -> (Graph n w) -> w

< weight   :: (Ix n,Num w) => n -> n -> (Graph n w) -> w

Adjacency matrix representation

> type Graph n w = Array (n,n) (Maybe w)
            
> mkGraph dir bnds@(l,u) es 
>     = emptyArray // ([((x1,x2),Just w) |(x1,x2,w)<-es] ++
>                      if dir then []
>                      else [((x2,x1),Just w) |(x1,x2,w)<-es,x1/=x2])
>       where
>       emptyArray
>           = array ((l,l),(u,u)) [((x1,x2),Nothing) | x1 <- range bnds, 
>                                                      x2 <- range bnds]

> adjacent g v1 = [ v2 | v2 <-nodes g,(g!(v1,v2))/= Nothing]

> nodes g       = range (l,u) where ((l,_),(u,_)) = bounds g

> edgeIn g (x,y)= (g!(x,y)) /= Nothing

modified from origional for debugging purposes...
< weight x y g  = w where (Just w) = g!(x,y)

> weight x y g  = case a of
>   (Just w) -> w
>   Nothing  -> error ("Cannot find weight for pair" ++ (show (x,y)))
>   where a = g!(x,y)

> edgesD g       = [(v1,v2,unwrap(g!(v1,v2))) 
>                       | v1 <-nodes g, v2 <- nodes g,
>                         edgeIn g (v1,v2)]
>     where unwrap (Just w) = w

> edgesU g       = [(v1,v2,unwrap(g!(v1,v2)))
>                    | v1 <-nodes g, v2 <- range (v1,u),
>                      edgeIn g (v1,v2)]
>     where (_,(u,_)) = bounds g          
>           unwrap (Just w) = w

***************************
* Algorithms for Graphs
***************************

Sparsity
--------

  This can be useful in picking a graph implementation, according to Rabhi and Lapalme..

> isSparse :: (Ix n, Num w) => ((Graph n w) -> [(n,n,w)]) -> Graph n w -> Bool
> isSparse edges g = ne < (nv * (log nv))
>   where ne  = len (edges g)
>         nv  = len (nodes g)
>         len = fromInteger.toInteger.length

> isSparseD :: (Ix a, Num b) => Graph a b -> Bool
> isSparseD = isSparse edgesD

> isSparseU :: (Ix a, Num b) => Graph a b -> Bool
> isSparseU = isSparse edgesU

Search
------

> depthFirstSearch start g = dfs [start] []
>  where
>    dfs [] vis    = vis
>    dfs (c:cs) vis 
>      | elem c vis = dfs cs vis
>      | otherwise  = dfs ((adjacent g c)++cs) (vis++[c])

depthFirstSearch' start g = reverse (dfs [start] [])
  where
   dfs [] vis     = vis
   dfs (c:cs) vis 
      | elem c vis = dfs cs vis
      | otherwise  = dfs ((adjacent g c)++cs) (c:vis)

depthFirstSearch'' start g = reverse (dfs (push start emptyStack) [])
 where
   dfs s vis 
    | (stackEmpty s)  = vis
    | elem (top s) vis = dfs (pop s) vis
    | otherwise       = let c = top s
                        in dfs (foldr push (pop s) (adjacent g c)) (c:vis)

breadthFirstSearch start g = reverse (bfs (enqueue start emptyQueue) [])
 where
  bfs q vis 
   | (queueEmpty q) = vis
   | elem (front q) vis = bfs (dequeue q) vis
   | otherwise  = let c = front q
                  in bfs (foldr enqueue (dequeue q) (adjacent g c))
                          (c:vis)

Find Paths
----------

  Given a set of alternatives, find all paths through that graph.

> findPaths :: [[n]] -> [[n]]
> findPaths []       = []
> findPaths (ns:[])  = map (\n -> [n]) ns
> findPaths (ns:nss) = concat (map (\n -> map (\ns -> n:ns) (findPaths nss)) ns)

> lowestWeightPath :: Ord o => ([n] -> o) -> [[n]] -> [n]
> lowestWeightPath cost paths = minimumBy f paths
>   where f a b = compare (cost a) (cost b)

> findLWPath :: Ord o => ([n] -> o) -> [[n]] -> [n]
> findLWPath cost = (lowestWeightPath cost) . findPaths

> highestWeightPath :: Ord o => ([n] -> o) -> [[n]] -> [n]
> highestWeightPath cost paths = maximumBy f paths
>   where f a b = compare (cost a) (cost b)

> findHWPath :: Ord o => ([n] -> o) -> [[n]] -> [n]
> findHWPath cost = (highestWeightPath cost) . findPaths

*********
* Tests
*********

> graphAM = mkGraph True (1,5) [(1,2,10),(1,3,20),(2,4,30),(3,4,40),(4,5,50)]

> testGraph = mkGraph False (1,8) [(1,4,1),(1,5,6),
>                                  (2,4,6),(2,5,0),
>                                  (3,4,3),(3,5,2),
>                                  (4,6,10),
>                                  (5,6,4),
>                                  (6,7,2),(6,8,3)]

> testAlternatives = [[1,2,3], [4,5], [6], [7,8]]

> testFindLWP = findLWPath cost testAlternatives
>   where cost []       = 0
>         cost (x:[])   = 0
>         cost (a:b:xs) = (weight a b testGraph) + (cost (b:xs))

g = mkGraph True (1,6) [(1,2,0),(1,3,0),(1,4,0), 
                        (3,6,0),(5,4,0),(6,2,0),
                        (6,5,0) ]
 
{-- End of Adjacency matrix representation --}

***************************
* Other Code / Impls
***************************

{-- Pointer representation --
data Graph n w = Vertex n [((Graph n w),w)] 

graphPTR = v1                   -- Figure 7.1 (c)
     where 
       v1 = Vertex 1 [(v2,12),(v3,34),(v5,78)]
       v2 = Vertex 2 [(v1,12),(v4,55),(v5,32)]
       v3 = Vertex 3 [(v1,34),(v4,61),(v5,44)]
       v4 = Vertex 4 [(v2,55),(v3,61),(v5,93)]
       v5 = Vertex 5 [(v1,78),(v2,32),(v3,44),(v4,93)]

-- End of Pointer representation --}

{-- Adjacency list representation --

type Graph n w = Array n [(n,w)]

mkGraph dir bnds es =
    accumArray (\xs x -> x:xs) [] bnds 
               ([(x1,(x2,w)) | (x1,x2,w) <- es] ++
                if dir then []
                else [(x2,(x1,w))|(x1,x2,w)<-es,x1/=x2])

adjacent g v    = map fst (g!v)

nodes g         = indices g

edgeIn g (x,y)  = elem y (adjacent g x)

weight x y g    = head [ c | (a,c)<-g!x , a==y]

edgesD g        = [(v1,v2,w) | v1<- nodes g , (v2,w) <-g!v1] 

edgesU g        = [(v1,v2,w) | v1<- nodes g , (v2,w) <-g!v1 , v1 < v2] 

graphAL = array (1,5) [(1,[(2,12),(3,34),(5,78)]),
                       (2,[(1,12),(4,55),(5,32)]),
                       (3,[(1,34),(4,61),(5,44)]),
                       (4,[(2,55),(3,61),(5,93)]),
                       (5,[(1,78),(2,32),(3,44),(4,93)])]

-- same graph but created with mkGraph but different ordering of arcs
graphAL' = mkGraph False (1,5) [(1,2,12),(1,3,34),(1,5,78),
                                (2,4,55),(2,5,32),
                                (3,4,61),(3,5,44),
                                (4,5,93)]
                          
-- End of Adjacency list representation --}

{-   Examples of evaluations and results

? graphAL
array (1, 5) [(1, [(2, 12), (3, 34), (5, 78)]), (2, [(1, 12), (4, 55), (5, 32)]), (3, [(1, 34), (4, 61), (5, 44)]), (4, [(2, 55), (3, 61), (5, 93)]), (5, [(1, 78), (2, 32), (3, 44), (4, 93)])]
? mkGraph (1,5) [(1,2,10),(1,3,20),(2,4,30),(3,4,40),(4,5,50)]
array (1, 5) [(1, [(3, 20), (2, 10)]), (2, [(4, 30)]), (3, [(4, 40)]), (4, [(5, 50)]), (5, [])]
? adjacent graphAL 4
[2, 3, 5]
? nodes graphAL
[1, 2, 3, 4, 5]
? edgesD graphAL
[(1, 2, 12), (1, 3, 34), (1, 5, 78), (2, 1, 12), (2, 4, 55), (2, 5, 32), (3, 1, 34), (3, 4, 61), (3, 5, 44), (4, 2, 55), (4, 3, 61), (4, 5, 93), (5, 1, 78), (5, 2, 32), (5, 3, 44), (5, 4, 93)]
? edgesU graphAL
[(1, 2, 12), (1, 3, 34), (1, 5, 78), (2, 4, 55), (2, 5, 32), (3, 4, 61), (3, 5, 44), (4, 5, 93)]
? edgeIn graphAL (3,2)
False
? edgeIn graphAL (3,4)
True
? weight 3 4 graphAL
61

? graphAM
array ((1, 1), (5, 5)) [((1, 1), Nothing), ((1, 2), Just 10), ((1, 3), Just 20), ((1, 4), Nothing), ((1, 5), Nothing), ((2, 1), Nothing), ((2, 2), Nothing), ((2, 3), Nothing), ((2, 4), Just 30), ((2, 5), Nothing), ((3, 1), Nothing), ((3, 2), Nothing), ((3, 3), Nothing), ((3, 4), Just 40), ((3, 5), Nothing), ((4, 1), Nothing), ((4, 2), Nothing), ((4, 3), Nothing), ((4, 4), Nothing), ((4, 5), Just 50), ((5, 1), Nothing), ((5, 2), Nothing), ((5, 3), Nothing), ((5, 4), Nothing), ((5, 5), Nothing)]
? adjacent graphAM 4
[5]
? nodes graphAM
[1, 2, 3, 4, 5]
? edgesD graphAM
[(1, 2, 10), (1, 3, 20), (2, 4, 30), (3, 4, 40), (4, 5, 50)]
? edgesU graphAM
[(1, 2, 10), (1, 3, 20), (2, 4, 30), (3, 4, 40), (4, 5, 50)]
? edgeIn graphAM (3,2)
False
? edgeIn graphAM (3,4)
True
? weight 3 4 graphAM
40

-}
