\section{The Performer Model}

The {\tt Performer} module implements a model of a performer's technical abilities and a persistence mechanism for that data.  It also includes a function for computing a difficulty score for a given piece of music, and a given performer.

\begin{verbatim}

> module Performer where
> import Ratio
> import IO
> import Array
> import Basics
> import ComposersWorkbench
> import Miscellany
> import SaxophoneFingerings

\end{verbatim}

\subsection{Representation}

  A {\tt Performer} has a name and description of abilities.  Currently, this discription is limited to the area of technical speed, but this representation could be expanded later to accommmodate areas such as intonation, atriculation, and improvisation.  The performer's technical speed is represented by a map of fingering transitions to {\tt FingeringTempo}s where the mapped tempo is the maximum tempo at which the performer has been observed to perform the transition successfully.  Each transition is also mapped to a confidence level which indicates the degree of confidence in the mapping between the transition and the tempo.  As more observations of a performer's performance of a given transition are input into the system, this confidence level is increased. 

\begin{verbatim}

> type Name = String
> type ConfidenceLevel = Integer -- from 1 to 10 
> type SaxFCombination 
>   = (SaxophoneFingeringNumber, SaxophoneFingeringNumber)
> type TechniqueMap 
>   = Array SaxophoneFingeringNumber 
>           (Array SaxophoneFingeringNumber 
>                  (FingeringTempo, ConfidenceLevel)) 

> data Abilities = Abilities TechniqueMap 
>  deriving (Read, Show, Eq)

> data Performer = Performer Name Abilities
>  deriving (Read, Show, Eq)

> maximumTempo :: TechniqueMap -> SaxFCombination -> FingeringTempo
> maximumTempo t s = fst ((t ! (minp s)) ! (maxp s)) 

> confidence :: TechniqueMap -> SaxFCombination -> ConfidenceLevel
> confidence t s = snd ((t ! (minp s)) ! (maxp s))

> updateRecord :: TechniqueMap -> SaxFCombination -> FingeringTempo -> 
>                 ConfidenceLevel -> TechniqueMap
> updateRecord t s f c = t//[(minp s, innerArray)]
>  where innerArray = (t ! (minp s))//[(maxp s,(f,c))]

> maxConfidence = 10
> lowConfidence = 5
> minConfidence = 1

> -- Max and Min for pairs
> maxp (a, b) = max a b
> minp (a, b) = min a b

\end{verbatim}

\subsection{Persistence}

Since the model of the performer will be updated iteratively, through interaction with the performer over long periods of time, it is necessary to persist the data to disk.   An additional data type and accompanying functions are provided for this purpose.  Haskell's {\tt Read} and {\tt Show} classes are used to implement this functionality, by reading and writing the data as text.

\begin{verbatim}

> data PerformerDB = PerformerDB FilePath Performer
>  deriving (Read, Show, Eq)

> -- An "initial" performer database with the "default" set of abilities
> makePDB :: FilePath -> Name -> PerformerDB
> makePDB f n = PerformerDB f (makePerformer n)

> makePerformer :: Name -> Performer 
> makePerformer n = Performer n (Abilities t)
>  where t     = array (1, 48) [(n, innerArray n) | n <- [1..48]]
>        innerArray x 
>          = array (x, 48) [(y, (tInit, cInit)) | y <- [x..48]]
>        tInit = 15
>        cInit = 2 

> -- I/O operations 
> loadDB :: FilePath -> IO PerformerDB
> loadDB f = do h <- openFile f ReadMode 
>               s <- hGetContents h
>               hClose h
>               return $ r s
>  where r :: String -> PerformerDB
>        r = read

> storeDB :: PerformerDB -> IO ()
> storeDB pdb@(PerformerDB f _) 
>  = do h <- openFile f WriteMode 
>       hPutStr h (show pdb)
>       hClose h

\end{verbatim}

\subsection{Difficulty Estimation}

One use of the performer database is to estimate, for any musical phrase, the difficulty of that phrase.  In the following, an algorithm is introduced for the computation of a ``difficulty measure'' for a given phrase and performer.  This estimate is also given a (composite) confidence level, or degree of accuracy.

This difficulty measure is a weighted average where fingering transitions in the phrase that are ``out of reach'' for a performer are given a high weight compared to transitions that are ``difficult but doable'' which recieve a lower weight, and transitions that are ``easy'', which are given no weight. Thus, a phrase whose transitions all are at speeds at or beneath the associated speeds in the {\tt TechniqueMap} will have a difficulty score of zero, while phrases that have some intervals whose speeds are within a small amount of the speeds associated in the {\tt TechniqueMap} will have a small difficulty measure, and phrases that include even a few transitions whose speeds are greater than a small amount more than the associated speeds in the {\tt TechniqueMap} will have a high difficulty measure.  

Where $s_p$ is the speed of a given transition in the phrase, $s_d$ is a speed of a given transition in the database, $b$ is a small number, $c$ is the number of ``close'' transitions in the phrase, for which $s_p < s_d + b \land s_p > s_d$, $f$ is the number of ``far'' transitions, for which $s_p >= s_d + b$, and $e$ is the number of ``easy'' transitions, for which $s_p <= s_d$, then the difficulty measure is 

\begin{displaymath}
\frac{c + 10 * f}{10 * (e + c + f)}.
\end{displaymath}

\begin{verbatim}

> type Difficulty = Double -- bounded from 0 to 1
> type DegreeOfConfidence = Double -- average from 1 to 10

> -- Takes a performer, a music, beats per minute, beats per measure, 
> --   and returns a difficulty rating.
> estimateDifficulty :: Performer -> Music -> Integer -> Integer -> 
>                       (Difficulty, DegreeOfConfidence)
> estimateDifficulty p@(Performer _ (Abilities t)) m bpm bpms = (d, c)
>   where c = (fromInteger $ sum $ map (confidence t) s) / 
>             (fromInt $ length s)
>         d :: Difficulty 
>         d = (fromInt ((nct + (10 * nft)))) /
>             (fromInt (10 * (net + nct + nft)))
>         s :: [SaxFCombination]
>         s = fingeringCombinations m
>         nct :: Int
>         nct = tmc (\(t',t'') -> (t' < (t'' + buffer)) && t' > t'')
>         nft = tmc (\(t',t'') -> t' >= (t'' + buffer)) 
>         net = tmc (\(t',t'') -> t' <= t'') 
>         tmc :: ((FingeringTempo, FingeringTempo) -> Bool) -> Int
>         tmc p = length $ filter p tm 
>         -- tm: tempo map (a,b) where a is T_Phrase and b is T_KB
>         tm :: [(FingeringTempo, FingeringTempo)] 
>         tm = zip (fingeringTempos m bpm bpms) $ map (maximumTempo t) s
>         buffer = 10 -- the amount by which we want the performer to "reach"

> fingeringCombinations :: Music -> [SaxFCombination]
> fingeringCombinations m 
>   = zipSeq $ map fingeringNumber $ findFingeringSequence m 

> -- Takes music, bpm, and beats per measure 
> fingeringTempos :: Music -> Integer -> Integer -> [FingeringTempo]
> fingeringTempos m bpm bpms 
>   = map (durToFT bpm bpms) $ take ((length pd) - 1) pd
>   where pd = extractDurations m -- durations of all notes in the phrase

> -- durToFT normalizes durations in the KB to fingering tempos. It 
> --   takes the bpm and bpMeas of a phrase and the duration of a note 
> --   and returns a tempo. It also rounds down the derived tempo to 
> --   the nearest integer.
> durToFT :: Integer -> Integer -> Dur -> FingeringTempo 
> durToFT bpMin bpMeasure dur 
>   = floor $ (fromInt (numerator r)) / (fromInt (denominator r))
>   where r = (fromInteger bpMin) / 
>             ((fromInteger 4) * dur * (fromInteger bpMeasure))

\end{verbatim}
